// inst.cc
// instrumentation code

#include "inst.h"     // this module
#include "mypush.h"   // traverseRoots
#include "gentrace.h" // trace
#include "mapfile.h"  // MapFile

#include <stdio.h>    // printf
#include <string.h>   // memset
#include <assert.h>   // assert
#include <stdlib.h>   // getenv

// --------------------- malloc hooks --------------------
// the global malloc, which is automatically inited to all-zeroes
Malloc instrumentationMalloc;
bool doInstrumenting = true;       // global switch for disabling hooks

// convenient alias
#define mal instrumentationMalloc


void *operator new(size_t sz)
{
  //printf("my new called\n");
  if (doInstrumenting) {
    return mal.malloc(sz);
  }
  else {
    return ::malloc(sz);
  }
}

void *operator new[](size_t sz)
{
  //printf("my new called\n");
  if (doInstrumenting) {
    return mal.malloc(sz);
  }
  else {
    return ::malloc(sz);
  }
}

void operator delete(void *ptr)
{
  //printf("my delete called\n");
  if (doInstrumenting) {
    return mal.free(ptr);
  }
  else {
    return ::free(ptr);
  }
}

void operator delete[](void *ptr)
{
  //printf("my delete[] called\n");
  if (doInstrumenting) {
    return mal.free(ptr);
  }
  else {
    return ::free(ptr);
  }
}


StackTrace globalTrace;         // fields inited to 0 automatically
enum { MAX_TRACE_LEN = 100 };


struct InDegree;      // fwd decl

// per-object data
struct NodeData {
  int inDegree;       // will be treated as an InDegree structure
  int trace[1];       // trace[0] is length of trace, trace[1..n] are values

  InDegree *getInDegree() { return (InDegree*)&inDegree; }
};


void perMallocInstrumenter(int *userData)
{
  //printf("perMallocInstrumenter called\n");

  static int frequency = -1;     // iters per call to analysis func
  static int iter = 0;           // iteration count

  iter++;

  // store a stack trace with each malloc
  if (globalTrace.trace == NULL) {
    // need to allocate space for the trace itself
    globalTrace.trace = (void const**)malloc(sizeof(void const*) * MAX_TRACE_LEN);
    globalTrace.length = MAX_TRACE_LEN;
  }

  #ifdef DO_TRACE
    // do the trace
    globalTrace.next = 0;
    trace(globalTrace);

    // allocate space to hold the trace, plus in-degree data
    NodeData *nodeData = (NodeData*)
      malloc(sizeof(NodeData) + globalTrace.next * sizeof(void*));

    // fill-in the structure
    nodeData->inDegree = 0;                 // in-degree algorithm expects this (?)
    nodeData->trace[0] = globalTrace.next;  // length of trace
    memcpy(nodeData->trace+1, globalTrace.trace, globalTrace.next * sizeof(void*));

    // store the pointer in the mymalloc header
    *((NodeData**)userData) = nodeData;
  #endif // DO_TRACE

  if (frequency == -1) {    // only true once per execution
    // read frequency from environment variable
    char *p = getenv("ITERS");
    if (p != NULL) {
      frequency = atoi(p);
    }
    else {
      frequency = 0;        // don't do it automatically
    }
  }

  if (frequency == 0) {
    return;      // never do anything here
  }

  if (iter % frequency == 0) {
    memoryGraphAnalysis();
  }
}


void perFreeInstrumenter(int *userData)
{
  #ifdef DO_TRACE
    // we need to free the trace structure
    free( *((NodeData**)userData) );
  #endif
}


// -------------------- graph edge analysis framework ----------------------
void *ptrRoundUp(void *p)
{
  return (void*) (((((int)p) + 3) / 4) * 4);
}

void *ptrRoundDown(void *p)
{
  return (void*) (((((int)p) + 0) / 4) * 4);
}


// In several places, we want a convenient way to name a node in the
// memory graph.  NULL means no node (where applicable), and heap objects
// are identified by their address.  To name the stack and the static data,
// we need distinguished codes.  These are those codes (they are arbitrary).
void * const stackNode = (void*)4;
void * const dataNode = (void*)8;


struct findTarget_data {
  void *code;      // code for pointed-at root, or alloc'd block, or NULL for none
  void *ptr;       // pointer in question
};

bool isPointerToRoot(void *start, void *end, bool isStack, void *extra)
{
  findTarget_data *dat = (findTarget_data*)extra;
  if (start <= dat->ptr &&
               dat->ptr < end) {
    dat->code = isStack? stackNode : dataNode;
    return false;    // found it
  }
  else {
    return true;     // keep looking
  }
}

void *findTarget(void *p)
{
  findTarget_data dat;
  dat.code = NULL;
  dat.ptr = p;

  // check the root set
  traverseRoots(isPointerToRoot, &dat);
  if (dat.code != NULL) {
    // found it
    return dat.code;
  }

  // check allocated memory; returns NULL or the required address
  return mal.findBase(p);
}


// user-supplied graph edge analysis fn
//   source   - ptr to source node, or stackNode or dataNode
//   target   - same as 'source', except the target of the edge
//   extra    - user-supplied arg
//   returns  - false to stop, true to continue
typedef bool (*EdgeAnalysis)(void *source, void *target, void *extra);

bool analyzeRegion(void *start, void *end, void *source,
                   EdgeAnalysis func, void *extra)
{
  start = ptrRoundUp(start);    /* next aligned address, if this isn't */
  end = ptrRoundDown(end);      /* prev aligned address, if this isn't */

  for (void **root = (void**)start; root < end; root++) {
    // extract the potential pointer at 'root' word
    void *p = *root;

    // see what it points at
    p = findTarget(p);

    // analyze it
    if (!func(source, p /*target*/, extra)) {
      return false;
    }
  }

  return true;    // wasn't interrupted
}

struct Analysis {
  EdgeAnalysis func;   // func to apply to each edge
  void *extra;         // extra params to 'func'
};

bool analyzeRoot(void *start, void *end, bool isStack, void *extra)
{
  Analysis *a = (Analysis*)extra;
  return analyzeRegion(start, end, isStack? stackNode : dataNode,
                       a->func, a->extra);
}

bool analyzeHeap(void *start, void *end, int*, void *extra)
{
  Analysis *a = (Analysis*)extra;
  return analyzeRegion(start, end, start /*source*/,
                       a->func, a->extra);
}

void analyzeMemory(EdgeAnalysis func, void *extra)
{
  Analysis a;
  a.func = func;
  a.extra = extra;
  if (traverseRoots(analyzeRoot, &a)) {
    mal.traverseMemory(analyzeHeap, &a);
  }
}


// -------------------------------------------------------------
// analysis: in-degree calculation
// record for how many things point to some object
struct InDegree {
  unsigned char fromStack;
  unsigned char fromData;
  unsigned char fromHeap;
  unsigned char fromSelf;

  bool operator== (InDegree const &obj) const {
    return fromStack == obj.fromStack &&
           fromData == obj.fromData &&
           fromHeap == obj.fromHeap &&
           fromSelf == obj.fromSelf;
  }

  InDegree& operator= (InDegree const &obj) {
    if (this != &obj) {
      fromStack = obj.fromStack;
      fromData = obj.fromData;
      fromHeap = obj.fromHeap;
      fromSelf = obj.fromSelf;
    }
    return *this;
  }

  void clear()
  {
    memset(this, 0, sizeof(this));
  }
};


#ifndef DO_TRACE
  inline InDegree *userDataToInDegree(int *userData)
  {
    return (InDegree*)userData;
  }
#else
  inline InDegree *userDataToInDegree(int *userData)
  {
    NodeData *data = *((NodeData**)userData);
    return data->getInDegree();
  }
#endif


struct countPointers_data {
  InDegree toStack;
  InDegree toData;
};

bool tallyPointer(void *source, void *target, void *extra)
{
  countPointers_data *dat = (countPointers_data*)extra;
  assert(source != NULL);

  // find records for the target
  InDegree *deg = NULL;
  if (target == NULL) {
    return true;    // nop, continue
  }
  else if (target == stackNode) {
    deg = &dat->toStack;
  }
  else if (target == dataNode) {
    deg = &dat->toData;
  }
  else {
    // points at a heap location
    deg = userDataToInDegree(mal.userData(target));
  }

  // find byte of interest
  unsigned char *b = NULL;
  if (source == stackNode) {
    b = &deg->fromStack;
  }
  else if (source == dataNode) {
    b = &deg->fromData;
  }
  else if (source == target) {
    b = &deg->fromSelf;
  }
  else {
    b = &deg->fromHeap;
  }

  // increment it, if possible (let things max out at 255)
  if (*b != 255) {
    (*b)++;
  }

  return true;     // continue
}


struct Histogram {
  enum { HISTSIZE=1024 };
  InDegree hist[HISTSIZE];     // block in-degree stats
  int count[HISTSIZE];         // # of blocks with those stats
  int next;                    // index of next unused spot

  void clear();
};

void Histogram::clear() 
{ 
  next = 0; 
  for (int i=0; i<HISTSIZE; i++) {
    count[i] = 0;
  }
}


bool createHistogram(void*, void*, int *userData, void *extra)
{
  Histogram &hist = *((Histogram*)extra);
  InDegree &deg = *userDataToInDegree(userData);

  // find histogram entry with identical stats
  int i;
  for (i=0; i<hist.next; i++) {
    if (hist.hist[i] == deg) {
      // found it
      hist.count[i] ++;
      return true;      // continue;
    }
  }

  // no entry has identical stats
  if (hist.next < hist.HISTSIZE) {
    // add a new entry
    hist.next++;
    hist.hist[i] = deg;
    hist.count[i] = 1;
  }
  else {
    // out of entries; pile them all into the last
    hist.count[i-1] ++;
  }

  return true;     // continue;
}

void printHistogram(Histogram const &hist)
{
  printf("count      fromStack   fromData   fromHeap  fromSelf\n");
  printf("--------   ---------   --------   --------  --------\n");
  for (int i=0; i<hist.next; i++) {
    InDegree const &deg = hist.hist[i];
    printf("%8d   %9d   %8d   %8d  %8d%c\n",
           hist.count[i],
           deg.fromStack, deg.fromData, deg.fromHeap, deg.fromSelf,
           i == hist.HISTSIZE? '*' : ' ');
  }
}


void printInDegree(char const *label, InDegree const &deg)
{
  printf("stack -> %s: %d\n", label, deg.fromStack);
  printf("data -> %s: %d\n", label, deg.fromData);
  printf("heap -> %s: %d\n", label, deg.fromHeap);
  printf("%s -> self: %d\n", label, deg.fromSelf);
}


bool clearCount(void*, void*, int *userData, void*)
{
  InDegree *indeg = userDataToInDegree(userData);
  assert(indeg);     // was getting segfaults, possibly related to inconsistent DO_TRACE
  indeg->clear();
  return true;    // continue;
}


Histogram histogram;

void inDegreeAnalysis()
{
  // step 1: clear current counts
  countPointers_data dat;
  memset(&dat, 0, sizeof(dat));
  mal.traverseMemory(clearCount, NULL);

  // step 2: walk memory, incrementing pointed-to count
  analyzeMemory(tallyPointer, &dat);

  // step 3: walk memory, using pointed-to count to construct histogram
  histogram.clear();
  mal.traverseMemory(createHistogram, &histogram);

  // step 4: output
  Malloc::Stats const &s = mal.getStats();
  printf("mallocs=%d, mallocdBytes=%d\n", s.mallocs, s.mallocdBytes);
  printf("frees=%d, freedBytes=%d\n", s.frees, s.freedBytes);
  printf("freeAllNodes=%d, freeAllBytes=%d\n", s.freeAllNodes, s.freeAllBytes);
  printf("leaks=%d, leakedBytes=%d\n", s.active(), s.activeBytes());
#if 0
  printHistogram("stack -> heap", hist.fromStack);
  printHistogram("data -> heap", hist.fromData);
  printHistogram("heap -> heap", hist.fromHeap);
#endif // 0
  printHistogram(histogram);
  printInDegree("stack", dat.toStack);
  printInDegree("data", dat.toData);
  printf("\n");
}


// ---------------------------------------------------------
// analysis: entire graph output
void printEndpoint(FILE *fp, void *pt)
{
  if (pt == stackNode) {
    fprintf(fp, "stack");
  }
  else if (pt == dataNode) {
    fprintf(fp, "data");
  }
  else {
    fprintf(fp, "%p", pt);
  }
}

void printEdge(FILE *fp, void *source, void *target)
{
  fputs("e ", fp);           // tag meaning "edge"
  printEndpoint(fp, source);
  fputc(' ', fp);
  printEndpoint(fp, target);
  fputc('\n', fp);
}

bool outputEdge(void *source, void *target, void *extra)
{
  if (target == NULL) {
    return true;    // skip; continue
  }

  // don't print stack->stack nor data->data
  if (source == target && (source == stackNode || source == dataNode)) {
    return true;    // don't print; continue
  }

  FILE *fp = (FILE*)extra;
  printEdge(fp, source, target);
  return true;      // continue
}


struct WriteNodeTraceInfo {
  FILE *fp;        // output file
  MapFile *map;    // maps addrs to names
};

bool writeNodeTrace(void *start, void *, int *userData,
                    void *extra)
{
  NodeData *data = *((NodeData**)userData);
  WriteNodeTraceInfo *info = (WriteNodeTraceInfo*)extra;

  fprintf(info->fp, "n %p ", start);     // n means "node"
  
  // start with i=3 to skip the known top 3 calls:
  //   perMallocInstrumentor, Malloc::malloc, __builtin_new
  enum { SKIP = 3 };

  // write the trace as a big string
  for (int i=SKIP; i < data->trace[0]; i++) {
    if (i != SKIP) {
      fputc('/', info->fp);         // separate the names
    }

    int addr = data->trace[i+1];    // +1 because [0] is the count
    if (info->map != NULL) {
      // map is available; use it
      char const *name = info->map->map((void*)addr);
      if (name == NULL) {
	name = "(addr to low)";
      }
      fputs(name, info->fp);
    }
    else {
      // map is not available; output raw addresses
      fprintf(info->fp, "%p", (void*)addr);
    }
  }
  fputc('\n', info->fp);

  return true;       // continue;
}


bool writeLeakEdges(void *start, void *, int *userData,
                    void *extra)
{
  FILE *fp = (FILE*)extra;
  InDegree *indeg = userDataToInDegree(userData);

  // is this node a leak node?
  if (indeg->fromStack == 0 &&
      indeg->fromData == 0 &&
      indeg->fromHeap == 0) {
    // yep; output a spurious edge so it shows up in applet
    printEdge(fp, start, start);
  }

  return true;   // continue
}


void graphOutput()
{
  // step 1: open a file
  char *fnamePrefix = getenv("GRAPH_PREFIX");
  if (!fnamePrefix) {
    return;     // don't output the graph
  }
  char fname[80];
  sprintf(fname, "%s.%d.g", fnamePrefix, mal.getStats().mallocs);
  FILE *fp = fopen(fname, "w");
  if (!fp) {
    fprintf(stderr, "failed to open graph file %s\n", fname);
    return;
  }

  // step 2: walk memory, outputting graph edges to file
  analyzeMemory(outputEdge, fp);

  // step 2b: walk allocated nodes, searching for leaks, and
  //          write "extra" self-edges for leaked nodes, so
  //          they show up in the Java applet
  mal.traverseMemory(writeLeakEdges, fp);
 
  #ifdef DO_TRACE
    // step 3: output per-node traces
    char const *mapFname = getenv("MAP_FNAME");
    if (mapFname != NULL) {
      MapFile map(mapFname);       // read map file
      WriteNodeTraceInfo info;
      info.fp = fp;
      info.map = &map;
      mal.traverseMemory(writeNodeTrace, &info);
    }
    else {
      WriteNodeTraceInfo info;
      info.fp = fp;
      info.map = NULL;
      mal.traverseMemory(writeNodeTrace, &info);
    }
  #endif // DO_TRACE

  // step 4: cleanup
  fclose(fp);
}


// -------------------- dispatcher -----------------
void memoryGraphAnalysis()
{
  //printf("memoryGraphAnalysis called\n");

  // during analysis, we want to be able to allocate without
  // disturbing the data
  doInstrumenting = false;

  try {
    inDegreeAnalysis();
    graphOutput();
  }
  catch (...) {
    fprintf(stderr, "Holy smokes!  Uncaught exception in analysis!\n");
  }

  // now, resume instrumentation
  doInstrumenting = true;
}
