// mymalloc.cc
// code for mymalloc.h

#include "mymalloc.h"      // this module

#include <assert.h>        // assert
#include <stdio.h>         // printf
#include <stdlib.h>        // malloc, free
#include <string.h>        // memset

//#define CHECKHEAP() checkHeap()
#define CHECKHEAP() ;

#if 0
// good ctor and dtor for "normal" uses
Malloc::Malloc()
  : first(NULL)
{
  memset(&stats, 0, sizeof(stats));
  CHECKHEAP();
}

Malloc::~Malloc()
{
  CHECKHEAP();
  if (!isEmpty()) {
    printf("warning: %d unfreed malloc blocks", numAllocated());
    freeAll();
  }
  delete index;
}

#else  // 0
// The primary intended use of this module is to instrument existing
// programs.  In this cases there will be one global instance, but we
// cannot (portably) predict the order in which the ctor and dtor
// will be called, relative to other initializations.  Therefore, we
// rely on the loader initializing the members to 0, and don't do
// anything in the dtor.

Malloc::Malloc() {}
Malloc::~Malloc() {}

#endif



void *Malloc::malloc(unsigned size)
{
  //assert(size > 0);
  // I turned off this because crypto++ violates it

  CHECKHEAP();
  stats.mallocs++;
  stats.mallocdBytes += size;

  // call normal system malloc
  void *mem = ::malloc(size + sizeof(Header));

  // write links, making this the first in the list
  Header *h = (Header*)mem;
  h->prev = NULL;
  h->next = first;
  h->size = size + sizeof(Header);
  h->userData = 0;

  // update old first's links
  if (first) {
    first->prev = h;
  }

  // and now we are the first
  first = h;

  // invalidate the index
  changed = true;

  // return ptr to user's space
  CHECKHEAP();
  return h->userStart();
}


void Malloc::free(void *ptr)
{
  CHECKHEAP();

  // get ptr to my header
  Header *h = getHeader(ptr);

  // update stats
  stats.frees++;
  stats.freedBytes += h->userBytes();

  // update 'first' ptr
  if (first == h) {
    first = first->next;
  }

  // patch this node out of the list
  if (h->prev) {
    h->prev->next = h->next;
  }
  if (h->next) {
    h->next->prev = h->prev;
  }

  // trash links so we can't be confused with a real block
  memset(h, 0, sizeof(*h));

  // finally, let real malloc get rid of this block
  ::free(h);
  CHECKHEAP();

  // invalidate the index
  changed = true;
}


// no attempt is made at optimization since realloc is silly anyway
void *Malloc::realloc(void *ptr, unsigned newSize)
{
  stats.reallocs++;

  // allocate a new block
  void *newBlock = malloc(newSize);

  // copy contents to preserve
  unsigned oldSize = allocatedSize(ptr);
  memcpy(newBlock, ptr, oldSize < newSize? oldSize : newSize);

  // free old block
  free(ptr);

  // invalidate the index
  changed = true;

  return newBlock;
}


// given a pointer that may point into an allocated block, or
// may point somewhere else entirely, return the header of that
// block (if the former), or NULL (if the latter)
void *Malloc::findBase(void const *ptr) const
{
  // to improve efficiency, build an index of object locations
  const_cast<Malloc*>(this)->      // it's const to the user
    buildIndex();

  Header *h = index->search(ptr);
  if (h != NULL) {
    return h->userStart();
  }
  else {
    return NULL;
  }


  #if 0     // old, slow code
    // iterate through all allocated blocks
    for (Header *h = first; h != NULL; h = h->next) {
      if (h->contains(ptr)) {
        return h->userStart();
      }
    }
    return NULL;
  #endif // 0
}


// verify my records of heap structure are ok;
// throw exception of they're bad
void Malloc::checkHeap() const
{
  // verify correct start; with this, and subsequent checks of
  // local correctness, an infinite loop is impossible
  assert(!first || first->prev == NULL);

  // iterate through all allocated blocks
  for (Header *h = first; h != NULL; h = h->next) {
    // verify it's ok
    assert(h->locallyOk());
  }

  // ok
}


void Malloc::freeAll()
{
  CHECKHEAP();
  while (first) {
    // remove first node from list
    Header *temp = first;
    assert(first->prev == NULL);
    first = first->next;
    if (first) {
      first->prev = NULL;
    }

    // update statistics
    stats.freeAllNodes++;
    stats.freeAllBytes += temp->userBytes();

    // release storage
    ::free(temp);
  }
  CHECKHEAP();

  // invalidate the index
  changed = true;
}


int Malloc::numAllocated() const
{
  int ct=0;
  for (Header *h = first; h != NULL; h = h->next) {
    ct++;
  }

  // if the statistics are working correctly, we have this information
  // already.. this isn't time-critical, though, so let's just test
  assert(ct == stats.active());

  return ct;
}


void Malloc::traverseMemory(Malloc::TraversalFunc func, void *extra)
{
  for (Header *h = first; h != NULL; h = h->next) {
    if (!func(h->userStart(), h->userEnd(), &h->userData, extra)) {
      break;
    }
  }

  // this traversal does not, by itself, invalidate the index
}


Malloc::Index::Index(int n)
  : numBlocks(n)
{
  array = new Header* [numBlocks];
}

Malloc::Index::~Index()
{
  delete[] array;
}


Malloc::Header *Malloc::Index::search(void const *addr) const
{
  // special-case an empty index
  if (numBlocks <= 0) {
    assert(numBlocks == 0);
    return NULL;
  }

  // bounds on the elements we're considering
  int low = 0;
  int high = numBlocks-1;

  // since we expect to test many words that are not pointers, let's
  // quickly rule-out addresses totally out of our range
  if (addr < array[low]->userStart() ||
      addr >= array[high]->userEnd()) {
    return NULL;
  }

  // binary search loop
  while (low < high) {
    // look at middle element
    int mid = (low+high)/2;
    Header *midHeader = array[mid];

    if (addr < midHeader->userStart()) {
      high = mid-1;
    }
    else if (addr >= midHeader->userEnd()) {
      low = mid+1;
    }
    else {
      // addr is between userStart and userEnd of 'mid'
      return midHeader;
    }
  }

  // either we've found it, or no object contains the address
  if (low == high) {        // else they crossed, definitely no hit
    Header *h = array[low];
    if (h->userStart() <= addr &&
                          addr < h->userEnd()) {
      return h;
    }
  }

  return NULL;
}


// treat a and b as pointers to integers, and compare them as such
// (they are really pointers to pointers, and we want to compare
// their pointer-values)
static int comparePointers(void const *a, void const *b)
{
  return *((int*)a) - *((int*)b);
}

void Malloc::buildIndex()
{
  if (index != NULL) {
    // if the index exists and is still valid, keep using it
    if (!changed) {
      // may as well check that things haven't changed
      assert(index->numBlocks == stats.active());
      return;
    }

    // since we're rebuilding, throw away the index
    delete index;
  }

  // allocate an index structure
  int numBlocks = numAllocated();
  index = new Index(numBlocks);
  Header **array = index->array;

  // walk memory, inserting blocks into the index
  int i = 0;
  for (Header *h = first; h != NULL; h = h->next, i++) {
    array[i] = h;
  }
  assert(i == numBlocks);

  // sort the index into increasing address order, to enable
  // binary search
  qsort(array, numBlocks, sizeof(Header*), comparePointers);

  // this index will be valid until the next change
  changed = false;
}


Malloc::Header const *Malloc::getHeaderC(void const *ptr) const
{
  Header const *h = ( (Header const*)ptr - 1 );
  assert(first && h->locallyOk());
  return h;
}


// this doesn't take 'void const *ptr' because the caller can modify
// the user-data member via the return value
int *Malloc::userData(void *ptr)
{
  CHECKHEAP();
  return &( getHeader(ptr)->userData );
}


unsigned Malloc::allocatedSize(void const *ptr) const
{
  CHECKHEAP();
  return getHeaderC(ptr)->userBytes();
}


bool Malloc::isBlock(void const *ptr) const
{
  CHECKHEAP();

  // get ptr to my header, check integrity
  Header const *h = ( (Header const*)ptr - 1 );
  return first && h->locallyOk();
}


void Malloc::Stats::print(FILE *fp) const
{
  fprintf(fp, "mallocs=%d, mallocdBytes=%d\n", mallocs, mallocdBytes);
  fprintf(fp, "frees=%d, freedBytes=%d\n", frees, freedBytes);
  fprintf(fp, "freeAllNodes=%d, freeAllBytes=%d\n", freeAllNodes, freeAllBytes);
  fprintf(fp, "active=%d, activeBytes=%d\n", active(), activeBytes());
  fprintf(fp, "reallocs=%d\n", reallocs);
}


// ------------------ test code ------------------------------
#ifdef MODULE_TEST

#ifndef __BORLANDC__
  #include <time.h>       // time()
  #define randomize() (srand(time(NULL)))
  #define random(num) (rand()%num)
#endif // !__BORLANDC__


void printStats(Malloc const &mal)
{
  printf("\n");
  mal.getStats().print(stdout);
}


bool checkSizes(void *start, void *end, int *userData, void *extra)
{
  int *traverseCt = (int*)extra;
  (*traverseCt)++;
  assert( (char*)end - (char*)start == *userData );
  return true;   // continue
}


// now I'm testing it in the usual configuration, where it is
// a global object (and therefore implicitly inited to 0)
Malloc mal;

void checkBase(void *p, int size, int addr, bool shouldEqual)
{
  void *ans = mal.findBase((void*)addr);
  if ((ans == p) != shouldEqual) {
    printf("p=%p, size=%d, addr=%p, shouldEqual=%d, ans=%p\n",
           p, size, (void*)addr, shouldEqual, ans);
  }
}


bool mymalloc_test(int argc, char *argv[])
{
  // quick test
  void *p1 = mal.malloc(10);
  void *p2 = mal.malloc(20);

  printStats(mal);

  assert(mal.findBase(p1) == p1);
  assert(mal.isBlock(p1) && mal.isBlock(p2));
  assert(mal.findBase((char*)p1 + 5) == p1);
  assert(mal.findBase((char*)p1 + 9) == p1);
  assert(mal.findBase((char*)p1 + 10) == NULL);
  assert(mal.findBase((char*)p1 - 1) == NULL);

  assert(mal.findBase((char*)p2 + 5) == p2);

  assert(mal.numAllocated() == 2);

  mal.free(p1);
  mal.free(p2);

  assert(mal.findBase(p1) == NULL);
  assert(mal.findBase(p2) == NULL);
  assert(!mal.isBlock(p1) && !mal.isBlock(p2));

  printStats(mal);

  assert(mal.isEmpty());
  assert(mal.numAllocated() == 0);

  // long, randomized test
  randomize();
  enum { PTRS = 100 };
  void *ptr[PTRS];
  memset(ptr, 0, sizeof(ptr));

  int iters = 2000;
  if (argc > 1) {
    iters = atoi(argv[1]);
  }

  for(int i=0; i<iters; i++) {
    // select an entry at random
    int entry = random(100);
    int newSize = random(1000);

    // test findBase
    if (ptr[entry] != NULL) {
      void *p = ptr[entry];
      int size = *(mal.userData(p));
      assert((unsigned)size == mal.allocatedSize(p));     // gcc sucks!
      assert(mal.isBlock(p));

      // isBlock sometimes segfaults on non-blocks
      //assert(!mal.isBlock(  (void*)( (int)p + 1 )  ) &&
      //       !mal.isBlock(  (void*)( (int)p - 1 )  ) );

      checkBase(p, size, (int)p - 1       , false);
      checkBase(p, size, (int)p           , size >= 1);
      checkBase(p, size, (int)p + 1       , size >= 2);
      checkBase(p, size, (int)p + size - 1, size >= 1);
      checkBase(p, size, (int)p + size    , false);
    }

    if (ptr[entry] == NULL) {
      // allocate
      ptr[entry] = mal.malloc(newSize);
      *(mal.userData(ptr[entry])) = newSize;   // for testing userData
    }
    else if (random(5) == 0) {
      // reallocate
      ptr[entry] = mal.realloc(ptr[entry], newSize);
      *(mal.userData(ptr[entry])) = newSize;
    }
    else {
      // deallocate
      mal.free(ptr[entry]);
      ptr[entry] = NULL;
    }
  }

  printStats(mal);

  // verify all alloc'd blocks have correct userData
  int traverseCt = 0;
  mal.traverseMemory(checkSizes, &traverseCt);
  printf("\n%d blocks traversed (all had correct userData)\n", traverseCt);
  traverseCt = 0;
  mal.traverseMemory(checkSizes, &traverseCt);
  printf("%d blocks traversed (all had correct userData)\n", traverseCt);

  mal.freeAll();
  printStats(mal);

  return true;
}


int main(int argc, char **argv)
{
  return mymalloc_test(argc, argv)? 0 : 2;
}

#endif // MODULE_TEST

